# Copyright (C) 2024- Davide Mollica <davide.mollica@inaf.it>
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of iactsim.
#
# iactsim is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# iactsim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with iactsim.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np

def pointing_dir(altitude: float, azimuth: float) -> np.ndarray:
    """Telescope pointing direction in the corsika-like local reference frame.

    *   North -> (0, 0) : x-axis
    *   East -> (0, 90)
    *   South -> (0, 180)
    *   West -> (0, 270) : y-axis
    *   Zenith -> (90, any) : z-axis

    Parameters
    ----------
    altitude : float
        Altitude angle in degrees.
    azimuth : float
        Azimuth angle in degrees. Measured from North, increasing towards East (clockwise).

    Returns
    -------
    direction : ndarray, shape (3,)
        Pointing unit vector in the local reference frame (North-East-Up).
    """
    azimuth_rad = np.deg2rad(azimuth)
    altitude_rad = np.deg2rad(altitude)
    return np.array([
        np.cos(altitude_rad) * np.cos(azimuth_rad),
        -np.cos(altitude_rad) * np.sin(azimuth_rad),
        np.sin(altitude_rad),
    ])

def local_to_pointing_rotation(altitude: float, azimuth: float) -> np.ndarray:
    """Compute the rotation matrix to transform from the local reference frame (North-East-Up) to the "pointing reference frame".

    This reference frame is defined such that when the telescope is
    pointing at the horizon towards the North (i.e. alt: 0, az: 0), the axes are aligned as follows:

    *   z-axis: aligned with the telescope's optical axis, pointing North.
    *   x-axis: perpendicular to the optical axis, pointing downward.
    *   y-axis: perpendicular to the optical axis, pointing west.

    Parameters
    ----------
    altitude : float
        Altitude angle of the telescope pointing, in degrees.
    azimuth : float
        Azimuth angle of the telescope pointing, in degrees. Measured from North, increasing towards East.

    Returns
    -------
    R : ndarray, shape (3, 3)
        Rotation matrix. When a vector in the local reference frame is multiplied by this matrix
        it is transformed into the telescope reference frame.

    Notes
    -----
    This is a simple rotation of the CORSIKA reference frame that move the up-vector (z) into the telescope direction.

    """
    theta = np.radians(90-altitude)
    phi = -np.radians(azimuth)

    ct = np.cos(theta)
    st = np.sin(theta)
    cp = np.cos(phi)
    sp = np.sin(phi)

    # pointing north (0,0)
    #  z axis is pointing north (local x axis)
    #  x axis is pointing down (opposite to local z axis)
    #  y axis is pointing west (local y axis)
    R = np.array([
        [ct*cp, ct*sp, -st],
        [  -sp,    cp,   0],
        [st*cp, st*sp,  ct]
    ])
    return R

# TODO: the default custom roration is the ASTRI one. Maybe this should be removed
def local_to_telescope_rotation(altitude: float, azimuth: float, custom_rotation=None) -> np.ndarray:
    """Compute the rotation matrix to transform from the local reference frame (North-West-Up) to the "telescope reference frame".

    The telescope reference frame is defined such that when the telescope is
    pointing at the horizon towards the North (i.e. alt: 0, az: 0), the axes are aligned as follows:

    *   z-axis: Aligned with the telescope's optical axis and pointing North.
    *   x-axis: Perpendicular to the optical axis, lying in the horizontal plane and pointing West.
    *   y-axis: Perpendicular to the optical axis, pointing upwards.

    Parameters
    ----------
    altitude : float
        Altitude angle of the telescope pointing, in degrees.
    azimuth : float
        Azimuth angle of the telescope pointing, in degrees. Measured from North, increasing towards East.

    Returns
    -------
    R : ndarray, shape (3, 3)
        Rotation matrix. When a vector in the local reference frame is multiplied by this matrix
        it is transformed into the telescope reference frame.
    """
    # Rotation into the local reference frame
    R = local_to_pointing_rotation(altitude, azimuth)

    # Rotation around the new z axis
    if custom_rotation is None:
        # Pointing north (0,0):
        #  z axis is pointing north (local x axis)
        #  x axis is pointing west (local y axis)
        #  y axis is pointing up (local z axis)
        custom_rotation = np.array([
            [ 0, 1, 0],
            [-1, 0, 0],
            [ 0, 0, 1]
        ])

    return custom_rotation @ R

def photon_to_local_rotation(a: np.ndarray) -> np.ndarray:
    """Compute the rotation matrix to transform from the photon reference frame to the local reference frame.

    The photon reference frame is defined as follows:
    ...TODO...
    
    Parameters
    ----------
    a : ndarray, shape (3,)
        Photon direction unit vector in the local reference frame.

    Returns
    -------
    R : ndarray, shape (3, 3)
        Rotation matrix. When a vector in the photon reference frame is multiplied by this matrix
        it is transformed into the local reference frame.
    """
    # Rotate `a` into [0,0,1]
    R = np.array([
        [1 + a[2] - a[0]**2, -a[0] * a[1], a[0] + a[0] * a[2]],
        [-a[0] * a[1], 1 + a[2] - a[1]**2, a[1] + a[1] * a[2]],
        [-a[0] - a[0] * a[2], -a[1] - a[1] * a[2], 1. + a[2] - a[0]**2 - a[1]**2]
    ]) / (1. + a[2])
    return R

def local_to_telescope_transform(
        ps: list | tuple | np.ndarray,
        vs: list | tuple | np.ndarray,
        altitude: float, azimuth: float,
        p0: list | tuple | np.ndarray
    ):
    """Transform positions and directions from the local reference frame 
    into the telescope reference frame given the telescope pointing direction and position.

    Parameters
    ----------
    ps : list, tuple, or numpy.ndarray
        Photon positions.
    vs : list, tuple, or numpy.ndarray
        Photon directions.
    altitude : float
        Telescope altitude angle in degree.
    azimut : float
        Telescope azimuth angle in degree.
    p0 : list, tuple, or numpy.ndarray
        Telescope position.
    
    Returns
    -------
    tel_ps : numpy.ndarray
        Array of transformed photon positions in the telescope coordinate system.
    tel_vs : numpy.ndarray
        Array of transformed photon direction vectors in the telescope coordinate system.

    """
    tel_ps = np.array(ps, copy=True)
    dtype = tel_ps.dtype
    tel_vs = np.array(vs, copy=True, dtype=dtype)
    tel_p0 = np.array(p0, copy=True, dtype=dtype)

    R = local_to_telescope_rotation(altitude, azimuth).astype(dtype)
    tel_ps -= tel_p0
    tel_ps = tel_ps @ R.T
    tel_vs = tel_vs @ R.T

    return tel_ps, tel_vs

def telescope_to_local_transform(
        ps: list | tuple | np.ndarray,
        vs: list | tuple | np.ndarray,
        altitude: float, azimuth: float,
        p0: list | tuple | np.ndarray
    ):
    """Transform positions and directions from the telescope reference frame 
    into the local reference frame given the telescope pointing direction and position.

    Parameters
    ----------
    ps : list, tuple, or numpy.ndarray
        Photon positions.
    vs : list, tuple, or numpy.ndarray
        Photon directions.
    altitude : float
        Telescope altitude angle in degree.
    azimut : float
        Telescope azimuth angle in degree.
    p0 : list, tuple, or numpy.ndarray
        Telescope position.
    
    Returns
    -------
    loc_ps : numpy.ndarray
        Array of transformed photon positions in the local coordinate system.
    loc_vs : numpy.ndarray
        Array of transformed photon direction vectors in the local coordinate system.

    """
    loc_ps = np.array(ps, copy=True)
    dtype = loc_ps.dtype
    loc_vs = np.array(vs, copy=True, dtype=dtype)
    loc_p0 = np.array(p0, copy=True, dtype=dtype)

    R = local_to_telescope_rotation(altitude, azimuth).T.astype(dtype)
    loc_ps = loc_ps @ R.T
    loc_vs = loc_vs @ R.T
    loc_ps += loc_p0

    return loc_ps, loc_vs

def moller_hughes_rotation(a: np.ndarray, b: np.ndarray) -> np.ndarray:
    """Moller-Hughes rotation matrix to rotate the versor `a` into a nearly parallel versor `b`.

    This function computes a rotation matrix that rotates a vector `a` into
    another vector `b`, where `a` and `b` are expected to be nearly parallel.
    It uses the Moller-Hughes algorithm, which is numerically stable for this
    specific case.

    Parameters
    ----------
    a : ndarray, shape (3,)
        The initial unit vector (versor) to be rotated.
    b : ndarray, shape (3,)
        The target unit vector (versor) onto which `a` should be rotated.

    Returns
    -------
    R : ndarray, shape (3, 3)
        Rotation matrix that describes the rotation transform: `R @ a` will be
        approximately equal to `b`.

    Raises
    ------
    RuntimeError
        If the algorithm fails to find a suitable rotation axis, which can happen
        if the input vectors are not nearly parallel or if they are not normalized.
    """
    # Check for normalization
    if not np.isclose(np.linalg.norm(a), 1.0) or not np.isclose(np.linalg.norm(b), 1.0):
        raise ValueError("Input vectors 'a' and 'b' must be unit vectors.")

    if (np.abs(a[0]) < np.abs(a[1])) and (np.abs(a[0]) < np.abs(a[2])):
        p = np.array([1., 0., 0.])
    elif (np.abs(a[1]) < np.abs(a[0])) and (np.abs(a[1]) < np.abs(a[2])):
        p = np.array([0., 1., 0.])
    # Maybe get rid of this since this function is always called for directions near to (0,0,-1)
    elif (np.abs(a[2]) < np.abs(a[0])) and (np.abs(a[2]) < np.abs(a[1])):
        p = np.array([0., 0., 1.])
    else:
        raise RuntimeError(
            f"Moller-Hughes algorithm failed to find a suitable rotation axis. "
            f"Input vectors: a = {a}, b = {b}. "
            f"This might be due to the vectors not being nearly parallel or an issue with the choice of 'p'."
        )

    u = p - a
    v = p - b
    u2 = np.dot(u, u)
    v2 = np.dot(v, v)
    uv = np.dot(u, v)
    R = np.identity(3) - 2. / u2 * np.outer(u, u) - 2. / v2 * np.outer(v, v) + 4. * uv / u2 / v2 * np.outer(v, u)
    return R


# Copyright (C) 2024- Davide Mollica <davide.mollica@inaf.it>
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of iactsim.
#
# iactsim is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# iactsim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with iactsim.  If not, see <https://www.gnu.org/licenses/>.

from tabulate import tabulate
import numpy as np
from IPython.display import display, HTML

from ._is_notebook import is_notebook

class BenchmarkTimer:
    """
    A class to time and manage benchmark results organized in "sections" and "measurements".

    Example
    -------

        .. code-block:: python

            from iactsim.utils import BenchmarkTimer
            import random
            import time

            class Foo():
                def __init__(self):
                    self.timer = BenchmarkTimer()
                    self.timer.add_section('foo1')
                    self.timer.add_section('foo2')

                def foo1(self):
                    t0 = time.time()
                    time.sleep(random.randint(0,50)/1000.)
                    t1 = time.time()
                    time.sleep(random.randint(0,5)/1000.)
                    t2 = time.time()
                    self.timer.add_entry('foo1', 'First', t1-t0)
                    self.timer.add_entry('foo1', 'Second', t2-t1)

                def foo2(self):
                    t0 = time.time()
                    time.sleep(random.randint(0,30)/1000.)
                    t1 = time.time()
                    self.timer.add_entry('foo2', 'foo2', t1-t0)
            
            foo = Foo()
            for i in range(100):
                foo.foo1()
                foo.foo2()
            foo.timer.print_results()
    
    """

    def __init__(self):
        """
        Initializes the BenchmarkTimer with an empty results dictionary.
        """
        self.results = {}  # {section_name: {measure_name: [times]}}
        self.active = True

    def add_section(self, section_name: str):
        """
        Adds a new section (e.g., a method name) to the results.

        Parameters
        ----------
        section_name : str
            The name of the section to add.
        """
        if section_name not in self.results:
            self.results[section_name] = {}

    def add_entry(self, section_name: str, measure_name: str, elapsed_time: float):
        """
        Adds a timing entry for a specific measure within a section.

        Parameters
        ----------
        section_name : str
            The name of the section.
        measure_name : str
            The name of the measurement within the section.
        elapsed_time : float
            The elapsed time in seconds.
        """
        if section_name not in self.results:
            raise(ValueError(f"Section {section_name} not found."))
        # Ensure measure exists
        if measure_name not in self.results[section_name]:
            self.results[section_name][measure_name] = []
        self.results[section_name][measure_name].append(elapsed_time)

    def get_results(self) -> dict:
        """
        Returns the raw results dictionary.

        Returns
        -------
        dict
            The results dictionary, structured as:
            {section_name: {measure_name: [time1, time2, ...]}}
        """
        return self.results

    def _calculate_totals(self):
        """
        Calculates and adds 'total_time' for each section.
        The 'total_time' will have an entry for each iteration, summing
        the times of all other measures *for that iteration*.
        """
        for section_name, section_data in self.results.items():
            # Find the maximum number of iterations
            max_iterations = 0
            for measure_name, times in section_data.items():
                if measure_name != "total_time":
                    max_iterations = max(max_iterations, len(times))

            # Initialize the total_time list
            if "total_time" not in self.results[section_name]:
                self.results[section_name]["total_time"] = []

            # Iterate through each iteration
            for i in range(max_iterations):
                total_time_for_iteration = 0
                # Sum the times for all measures in this iteration
                for measure_name, times in section_data.items():
                    if measure_name != "total_time" and i < len(times):
                        total_time_for_iteration += times[i]

                # Add/update the total time for this iteration
                if i < len(self.results[section_name]["total_time"]):
                  self.results[section_name]["total_time"][i] = total_time_for_iteration
                else:
                  self.results[section_name]["total_time"].append(total_time_for_iteration)

    def _prepare_data(self, compute_total: bool) -> list[list]:
        """
        Prepares the data for printing (calculates stats and overall total).

        Returns
        -------
        list[list]
            A list of lists, where each inner list represents a row
            in the table, ready for formatting.
        """
        self._calculate_totals()
        data = []

        for section_name, section_data in self.results.items():
            for measure_name, times in section_data.items():
                n_calls = len(times)
                if n_calls > 0:
                    times_np = np.array(times) * 1000  # Convert to milliseconds
                    min_time = np.min(times_np)
                    max_time = np.max(times_np)
                    mean_time = np.mean(times_np)
                    median_time = np.median(times_np)
                    stdev_time = np.std(times_np) if len(times_np) > 1 else 0.0
                else:
                    times_np = 0
                    min_time = 0
                    max_time = 0
                    mean_time = 0
                    median_time = 0
                    stdev_time = 0
                data.append([
                    section_name,
                    measure_name,
                    "{}".format(n_calls),
                    "{:.3f}".format(mean_time),
                    "{:.3f}".format(median_time),
                    "{:.3f}".format(stdev_time),
                    "{:.3f}".format(min_time),
                    "{:.3f}".format(max_time),
                ])

        # Calculate overall total (sum of all section totals) if every part is called n times
        # number_of_calls = [len(self.results[section]['total_time']) for section in self.results.keys()]
        # if len(set(number_of_calls)) == 1 and compute_total:
        #     totals = np.asarray([self.results[section]['total_time'] for section in self.results.keys()]).sum(axis=0) * 1000
        #     min_time = np.min(totals)
        #     max_time = np.max(totals)
        #     mean_time = np.mean(totals)
        #     median_time = np.median(totals)
        #     stdev_time = np.std(totals) if len(totals) > 1 else 0.0

            # data.append([]) # add an empty row
            # data.append([
            #     "Total",
            #     "",
            #     "{:.3f}".format(min_time),
            #     "{:.3f}".format(max_time),
            #     "{:.3f}".format(mean_time),
            #     "{:.3f}".format(median_time),
            #     "{:.3f}".format(stdev_time)
            # ])

        return data

    def _build_html_table(self, data: list[list]) -> str:
        """
        Builds the HTML table string for Jupyter Notebooks.  Internal method.

        Parameters
        ----------
        data : list[list]
            The prepared data (list of rows).

        Returns
        -------
        str
            The complete HTML table as a string.
        """
        headers = ["Section", "Measure", "N calls", "Mean (ms)", "Median (ms)", "Stdev (ms)", "Min (ms)", "Max (ms)"]
        html_output = "<table>\n"
        html_output += "<tr>" + "".join(f"<th>{header}</th>" for header in headers) + "</tr>\n"

        for row_data in data:
            if not row_data:  # Check for empty row
                html_output += "<tr>" + "".join("<td></td>" for _ in headers) + "</tr>\n" #empty row
            else:
                measure_name = row_data[1]
                # Apply bold formatting to total_time
                if measure_name == "total_time" or row_data[0] == "Total":
                    row_data = [f"<b>{x}</b>" for x in row_data]
                html_output += "<tr>" + "".join(f"<td>{value}</td>" for value in row_data) + "</tr>\n"


        html_output += "</table>"
        return html_output

    def _print_console_table(self, data: list[list], return_str=False):
        """
        Prints the formatted table to the console. Internal method.

        Parameters
        ----------
        data : list[list]
            The prepared data (list of rows).
        """
        headers = ["Section", "Measure", "N calls", "Mean (ms)", "Median (ms)", "Stdev (ms)", "Min (ms)", "Max (ms)"]

        formatted_data = []
        for row in data:
            if not row:
                formatted_data.append([""] * len(headers)) #empty row
            elif row[1] == "total_time" and not return_str:  #bold in console
                formatted_row = [f"\033[1m{x}\033[0m" for x in row]
                formatted_data.append(formatted_row)
            else:
                formatted_data.append(row)
        
        table = tabulate(formatted_data, headers=headers, tablefmt="grid")

        if return_str:
            return table
        else:
            print(table)


    def print_results(self, title: str = None, show_total: bool = True, return_str: bool = False):
        """
        Prints the timings, handling both notebook and console output.

        Parameters
        ----------
        title : str, optional
            An optional title to display above the table.
        show_total : bool, optional
            Add a last row with the sum of all sections.
        """
        if not self.active:
            return
        
        data = self._prepare_data(compute_total=show_total)

        if is_notebook() and not return_str:
            if title is not None:
                display(HTML(f'<h2>{title}</h2>'))
            html_table = self._build_html_table(data)
            display(HTML(html_table))
        else:
            if title is not None and not return_str:
                print(title.center(80))
            return self._print_console_table(data, return_str=return_str)

    def clear(self):
        """Resets the timer, keeping sections and measures, but clearing samples."""
        for section_data in self.results.values():
            for measure_name in section_data:
                section_data[measure_name] = []


import numpy as np

from ._surface import SurfaceShape
from ._surface_misc import SurfaceProperties

from ..visualization._iactsim_style import iactsim_style

class FresnelSurfacePropertiesGenerator:
    """
    Handles the calculation and population of SurfaceProperties objects 
    based on Fresnel equations for a given Surface instance.
    """

    @staticmethod
    def _calculate_fresnel(n1, n2, theta_deg, polarization='unpolarized'):
        """
        Perform a vectorized Fresnel calculation for a specific polarization.

        Parameters
        ----------
        n1 : np.ndarray
            Refractive index of the incident medium. Shape must be broadcastable 
            to (N_angles, N_wl).
        n2 : np.ndarray
            Refractive index of the transmission medium. Shape must be broadcastable 
            to (N_angles, N_wl).
        theta_deg : np.ndarray
            Incident angles in degrees. Shape must be broadcastable 
            to (N_angles, N_wl).
        polarization : str, optional
            The polarization mode to compute. Options are:
            - 'unpolarized': Average of s- and p-polarization.
            - 's', 'rs': s-polarization only.
            - 'p', 'rp': p-polarization only.
            Default is 'unpolarized'.

        Returns
        -------
        np.ndarray
            Reflectance matrix R with shape (N_angles, N_wl).

        Raises
        ------
        ValueError
            If an invalid polarization string is provided.
        """
        theta_i = np.radians(theta_deg)
        
        # Snell's law
        sin_theta_t = (n1 / n2) * np.sin(theta_i)

        # Total internal reflection
        tir_mask = np.abs(sin_theta_t) > 1.0

        sin_theta_t_safe = np.where(tir_mask, 0.0, sin_theta_t)
        theta_t = np.arcsin(sin_theta_t_safe)

        # Amplitude coefficients
        cos_theta_i = np.cos(theta_i)
        cos_theta_t = np.cos(theta_t)

        # Initialize result
        R_out = None

        # Calculate s-pol if needed
        if polarization in ['s', 'rs', 'unpolarized']:
            rs_num = n1 * cos_theta_i - n2 * cos_theta_t
            rs_den = n1 * cos_theta_i + n2 * cos_theta_t
            rs = np.divide(rs_num, rs_den, out=np.zeros_like(rs_num), where=rs_den!=0)
            Rs = np.abs(rs)**2
            Rs = np.where(tir_mask, 1.0, Rs)
            
            if polarization in ['s', 'rs']:
                R_out = Rs

        # Calculate p-pol if needed
        if polarization in ['p', 'rp', 'unpolarized']:
            rp_num = n2 * cos_theta_i - n1 * cos_theta_t
            rp_den = n2 * cos_theta_i + n1 * cos_theta_t
            rp = np.divide(rp_num, rp_den, out=np.zeros_like(rp_num), where=rp_den!=0)
            Rp = np.abs(rp)**2
            Rp = np.where(tir_mask, 1.0, Rp)

            if polarization in ['p', 'rp']:
                R_out = Rp

        # Unpolarized
        if polarization == 'unpolarized':
            R_out = 0.5 * (Rs + Rp)

        if R_out is None:
            raise ValueError(f"Invalid polarization mode: '{polarization}'. "
                             "Use 's', 'p', or 'unpolarized'.")

        return R_out

    def generate(self, surface, wavelengths=None, angles=None, polarization='unpolarized', inplace=False):
        """
        Generate a populated SurfaceProperties object.

        Parameters
        ----------
        surface : iactsim.optics.Surface
            The surface object containing `material_in` and `material_out` attributes,
            which must have `get_refractive_index(wavelengths)` methods.
        wavelengths : np.ndarray, optional
            1D array of wavelengths in nm. If None, defaults to `np.arange(200, 901, 1)`.
        angles : np.ndarray, optional
            1D array of incidence angles in degrees. If None, defaults to `np.arange(0, 91, 1)`.
        polarization : {'unpolarized', 's', 'p', 'rs', 'rp'}, optional
            The specific polarization to compute for reflectance and transmittance.
            Default is 'unpolarized'.
        inplace : bool, optional
            If True, modifies the provided surface properties in place.
            Default is False.

        Returns
        -------
        iactsim.optics.SurfaceProperties
            If `inplace` is False, a populated properties object containing the calculated matrices:
            - `reflectance`, `transmittance` (Front interface)
            - `reflectance_back`, `transmittance_back` (Back interface)
            - `wavelength`, `incidence_angle` vectors.
        """
        pol_clean = polarization.lower().strip()

        # Apply defaults
        if wavelengths is None:
            wavelengths = np.arange(200, 901, 1)
        
        if angles is None:
            angles = np.arange(0, 91, 1)

        # Ensure inputs are numpy arrays
        wl = np.asarray(wavelengths)
        ang = np.asarray(angles)

        # Retrieve materials and reshape for broadcasting
        # n: (1, N_wl) | ang: (N_ang, 1)
        n_in = surface.material_in.get_refractive_index(wl).reshape(1, -1)
        n_out = surface.material_out.get_refractive_index(wl).reshape(1, -1)
        ang_reshaped = ang.reshape(-1, 1)

        # For a cylindrical solid: n_in is the material inside; n_out is the material outside.
        if surface._shape == SurfaceShape.CYLINDRICAL:
            n_in, n_out = n_out, n_in

        # Front interface (n_in -> n_out)
        R_front = self._calculate_fresnel(n_in, n_out, ang_reshaped, pol_clean)
        
        # Cleanup
        nan_mask_f = np.isnan(R_front)
        T_front = 1.0 - R_front

        # NaNs will be treated as absorption
        R_front[nan_mask_f] = 0.0
        T_front[nan_mask_f] = 0.0

        # Back interface (n_out -> n_in)
        R_back = self._calculate_fresnel(n_out, n_in, ang_reshaped, pol_clean)
        
        # Cleanup
        nan_mask_b = np.isnan(R_back)
        T_back = 1.0 - R_back

        # NaNs will be treated as absorption
        R_back[nan_mask_b] = 0.0
        T_back[nan_mask_b] = 0.0

        # Populate Object
        surface_prop = SurfaceProperties()
        surface_prop.wavelength = wl
        surface_prop.incidence_angle = ang
        
        surface_prop.reflectance = R_front
        surface_prop.transmittance = T_front
        surface_prop.reflectance_back = R_back
        surface_prop.transmittance_back = T_back

        if inplace:
            surface.properties = surface_prop
        else:
            return surface_prop


# Copyright (C) 2024- Davide Mollica <davide.mollica@inaf.it>
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of iactsim.
#
# iactsim is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# iactsim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with iactsim.  If not, see <https://www.gnu.org/licenses/>.

from pathlib import Path as _Path
from typing import (
    Tuple as _Tuple,
    List as _List,
    Union as _Union,
    Dict as _Dict,
    Any as _Any
)

import numpy as _np
import yaml as _yaml

from ...io._path_loader import PathLoader as _PathLoader


from ._os import AstriOpticalSystem
from ._camera import AstriCameraGeometry, AstriCherenkovCamera
from ..._iact import IACT as _IACT

class AstriTelescope(_IACT):
    """ASTRI telescope.

    See Also
    --------
        :py:class:`iactsim.IACT`:
    """
    def __init__(
        self,
        optical_system = None,
        camera: AstriCherenkovCamera = None,
        position: _Tuple[float, float, float] | _List[float] | _np.ndarray = (0.,0.,0.),
        pointing: _Tuple[float, float] | _List[float] | _np.ndarray = (0,0),
    ):
        if optical_system is None:
            optical_system = AstriOpticalSystem()
        if camera is None:
            camera = AstriCherenkovCamera()
        super().__init__(
            optical_system = optical_system,
        )
        self.position = position
        self.pointing = pointing
        self.camera = camera
    
    def load_atmospheric_transmission(self, filepath: str):
        wavelength, height, value = self.optical_system.load_efficiency_data(filepath)
        self.atmospheric_transmission.altitude = height
        self.atmospheric_transmission.wavelength = wavelength
        self.atmospheric_transmission.value = value

    def configure(self, config: _Union[str, _Path, _Dict[str, _Any]]) -> None:
        """Configure the telescope with a yaml configuration file or dictionaries.
        The configuration files can contain more than a document to initialize also
        the optical system and the Cherenkov camera::

            ----
            # Telescope configuration
            # ...
            # ...
            ----
            # Optical system configuration
            # ...
            # ...
            ----
            # Camera configuration
            # ...
            # ...

        Parameters
        ----------
        config : str, path-like, or dict
            Path to the configuration file (str or pathlib.Path),
            or a dictionary containing the configuration.
        """
        configs = [None]*3

        if isinstance(config, (str, _Path)):
            with open(config, 'r') as file:
                tel_conf = _yaml.load_all(file, Loader=_PathLoader)
                for i,d in enumerate(tel_conf):
                    configs[i] = d
        elif isinstance(config, list):
            for i,d in enumerate(config):
                configs[i] = d
        elif isinstance(config, dict):
            configs[0] = config
        else:
            raise TypeError(
                "config must be a string (path), pathlib.Path, a list of dictionaries or a dictionary, "
                f"not {type(config)}"
            )

        n_documents = sum(d is not None for d in configs)

        if n_documents > 3:
            raise(RuntimeError('The configuration file can contain at most 3 document in the following order: telescope, optical system and camera.'))

        if n_documents == 0:
            raise(RuntimeError('The configuration file is empty.'))
        
        self._configure(configs[0])

        if configs[1] is not None:
            self.optical_system.configure(configs[1])
        
        if configs[2] is not None:
            self.camera.configure(configs[2])
        
        self.cuda_init()

    def _configure(self, tel_conf: _Dict[str, _Any]) -> None:
        """Configure the telescope with a dictionary.
        """
        # Assign only these
        assign_directly = [
            'altitude',
            'azimuth',
            'show_progress'
        ]

        for name in tel_conf.keys():
            value = tel_conf[name]
            if value is not None:
                if name in assign_directly:
                    setattr(self, name, value)
                if name == 'atmospheric_transmission':
                    self.load_atmospheric_transmission(value)


# Copyright (C) 2024- Davide Mollica <davide.mollica@inaf.it>
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of iactsim.
#
# iactsim is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# iactsim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with iactsim.  If not, see <https://www.gnu.org/licenses/>.

from dataclasses import dataclass
import numpy as np
from tabulate import tabulate
from typing import Tuple, Optional, Any
from numpy.typing import ArrayLike

from ._sag import sag
from ._surface_misc import SurfaceType, ApertureShape, SurfaceProperties, SurfaceShape, SensorSurfaceType
from ._materials import Material, Materials

@dataclass
class SurfaceVisualProperties:
    """
    Stores visualization attributes for an optical surface.

    Parameters
    ----------
    color : tuple of float, optional
        RGB color tuple (0.0-1.0). If None, the visualizer uses default logic based on surface type.
    opacity : float, optional
        Opacity (0.0 transparent - 1.0 opaque). If None, visualizer uses default logic based on surface type.
    specular : float, optional
        Specular reflection intensity (0.0-1.0).
    wireframe : bool, optional
        If True, forces wireframe representation for this surface. Default is False.
    resolution : float, optional
        Mesh resolution in mm. If None, uses the visualizer logic based on the surface extent.
    visible : bool, optional
        Whether the surface is visible in the render. Default is True.
    """
    color: Optional[Tuple[float, float, float]] = None
    opacity: Optional[float] = None
    specular: Optional[float] = None
    wireframe: bool = False
    resolution: Optional[float] = None
    visible: bool = True

class Surface:
    """Represents an optical surface.

        Parameters
        ----------
        type : SurfaceType
            Type of surface (reflective, refractive, opaque, dummy)
        position : array-like, optional
            Position of the surface center [x, y, z], by default None
        tilt_angles : array-like, optional
            Tilt angles (clockwise x-tilt, y-tilt and z_tild) in degrees, by default no tilt.
        scattering_dispersion : float, optional
            Standard deviation of Gaussian scattering after reflection/transmission, by default 0.0
        properties : SurfaceProperties, optional
            Reflectance/transmittance properties of the surface, by default None
        material_in : material, optional
            Material of the space before the surface front side. I.e. the material in which
            photons arriving with a negative z-component of their direction vector are travelling, 
            by default standard air.
        material_out : material, optional
            Material of the space after the surface back side. I.e. the material in which
            photons arriving with a positive z-component of their direction vector are travelling, 
            by default standard air.
        visual_properties : SurfaceVisualProperties, optional
            Visual properties of the surface, by default None
        name : str, optional
            Surface name
    """

    def __init__(self,
                 surface_type: SurfaceType,
                 position: ArrayLike = np.array([0., 0., 0.]),
                 tilt_angles: ArrayLike = np.array([0., 0., 0.]),
                 scattering_dispersion: float = 0.0,
                 properties: Optional[SurfaceProperties] = None,
                 material_in: Material = Materials.AIR,
                 material_out: Material = Materials.AIR,
                 visual_properties: Optional['SurfaceVisualProperties'] = None,
                 name: Optional[str] = None):
        self.type = surface_type
        self.position = position
        self.tilt_angles = tilt_angles
        self.scattering_dispersion = scattering_dispersion
        self.properties = properties if properties is not None else SurfaceProperties()
        self.material_in = material_in
        self.material_out = material_out
        self.name = name
        self.sensor_type = SensorSurfaceType.NONE
        if visual_properties is not None:
            self.visual_properties = visual_properties
        else:
            self.visual_properties = SurfaceVisualProperties()

        self._custom_rotation_matrix = None

    @property
    def axial_position(self) -> float:
        return self._axial_position

    @axial_position.setter
    def axial_position(self, value: float):
        self._axial_position = value
        self._position[-1] = value

    @property
    def position(self) -> np.ndarray:
        return self._position

    @position.setter
    def position(self, value: np.ndarray):
        self._position = np.asarray(value)
        self._axial_position = value[-1]
    
    def get_rotation_matrix(self) -> np.ndarray:
        """
        Calculates the 3D rotation matrix to transform into the surface reference system.
        The computation is based on surface tilt angles and assumes rotations 
        are applied in the order (last first): Z @ Y @ X. Each rotation and follow the right-hand rule. 
        Tilt angles are expected in degrees.

        A custom rotation matrix will be used if it has been set using :py:meth:`set_rotation_matrix` method.
        
        Returns
        -------
            np.ndarray: A 3x3 rotation matrix.
        """
        # Return the custom rotation matrix if has been defined
        if self._custom_rotation_matrix is not None:
            return self._custom_rotation_matrix.reshape(3,3)

        x_tilt_rad, y_tilt_rad, z_tilt_rad = np.deg2rad(self.tilt_angles)
        cos_x = np.cos(x_tilt_rad)
        sin_x = np.sin(x_tilt_rad)
        cos_y = np.cos(y_tilt_rad)
        sin_y = np.sin(y_tilt_rad)
        cos_z = np.cos(z_tilt_rad)
        sin_z = np.sin(z_tilt_rad)

        x_rot = np.array([
            [1.0,    0.,     0.],
            [ 0., cos_x, -sin_x],
            [ 0., sin_x,  cos_x]
        ])

        y_rot = np.array([
            [ cos_y,  0., sin_y],
            [    0.,  1.,    0.],
            [-sin_y,  0., cos_y]
        ])

        z_rot = np.array([
            [ cos_z, -sin_z, 0.],
            [ sin_z,  cos_z, 0.],
            [    0.,     0., 1.],
        ])

        return  (z_rot @ y_rot @ x_rot).T
    
    def set_rotation_matrix(self, a_matrix):
        """Set a custom rotation matrix instead of deriving it from tilt angles.
        To use tilt angles set it to ``None``.

        Parameters
        ----------
        a_matrix : np.ndarray
            3D rotation matrix to transform from telescope to surface reference frame.
        """
        self._custom_rotation_matrix = a_matrix
    
    def __repr__(self) -> str:
        """Returns a string representation of the surface object."""
        return f"{self.__class__.__name__}(half_aperture={self.type}, ...)"
    
    def _table_data(self) -> list[list[Any]]:
        """Helper function to generate data for table representation."""
        table = []
        if self.name is not None:
            table.append(["Name", self.name])
        table += [
            ["Surface Type", self.type],
            ["Position", ', '.join([f'{x:.2f}' for x in self.position])],
            ["Tilt Angles", ', '.join([f'{x:.2f}' for x in self.tilt_angles])],
            ["Scattering Sigma", self.scattering_dispersion],
            ["Properties wavelegth", "set" if self.properties.wavelength is not None else "not set"],
            ["Properties incidence angle", "set" if self.properties.incidence_angle is not None else "not set"],
            ["Transmittance value", "set" if self.properties.transmittance is not None else "not set"],
            ["Reflectance value", "set" if self.properties.reflectance is not None else "not set"],
            ["Absorption value", "set" if self.properties.absorption is not None else "not set"],
        ]
        return table

    def __str__(self) -> str:
        """Returns a string representation of the surface object formatted as a table."""
        return tabulate(self._table_data(), headers=["Attribute", "Value"], tablefmt="fancy_grid")
    
    def _repr_markdown_(self) -> str:
        """Returns a Markdown representation of the surface object."""
        return tabulate(self._table_data(), headers=["Attribute", "Value"], tablefmt="pipe")

    def _repr_html_(self) -> str:
        """Returns an HTML representation of the surface object."""
        return tabulate(self._table_data(), headers=["Attribute", "Value"], tablefmt="html")

class AsphericalSurface(Surface):
    """Represents an aspherical optical surface.

        Parameters
        ----------
        half_aperture : float
            Half-aperture of the surface.
        curvature : float
            Surface curvature (1/radius of curvature)
        conic_constant : float
            Conic constant
        aspheric_coefficients : array-like
            Array of aspheric coefficients
        central_hole_half_aperture : float, optional
            Half-aperture of the central hole (0.0 if no hole), by default 0.0
        aperture_shape : ApertureShape, optional
            Shape of the aperture (circular, hexagonal or rectangular), by default ApertureShape.CIRCULAR
        central_hole_shape : ApertureShape, optional
            Shape of the central hole (circular, hexagonal or rectangular), by default ApertureShape.CIRCULAR
        is_fresnel : bool, optional
            Whether the surface is an ideal Fresnel lens/mirror, by default False
        surface_type : SurfaceType, optional
            Type of surface (reflective, refractive, opaque, dummy), by default SurfaceType.REFLECTIVE
        position : array-like, optional
            Position of the surface center [x, y, z], by default None
        tilt_angles : array-like, optional
            Tilt angles (clockwise x-tilt, y-tilt and z_tild) in degrees, by default no tilt.
        scattering_dispersion : float, optional
            Standard deviation of Gaussian scattering after reflection/transmission, by default 0.0
        properties : SurfaceProperties, optional
            Reflectance/transmittance properties of the surface, by default None
        material_in : material, optional
            Material of the space before the surface front side. I.e. the material in which
            photons arriving with a negative z-component of their direction vector are travelling, 
            by default standard air.
        material_out : material, optional
            Material of the space after the surface back side. I.e. the material in which
            photons arriving with a positive z-component of their direction vector are travelling, 
            by default standard air.
        name : str, optional
            Surface name
    """

    def __init__(self,
                 half_aperture: float,
                 curvature: float,
                 conic_constant: float,
                 aspheric_coefficients: ArrayLike,
                 central_hole_half_aperture: float = 0.0,
                 aperture_shape: ApertureShape = ApertureShape.CIRCULAR,
                 central_hole_shape: ApertureShape = ApertureShape.CIRCULAR,
                 is_fresnel: bool = False,
                 surface_type: SurfaceType = SurfaceType.REFLECTIVE,
                 position: ArrayLike = np.array([0., 0., 0.]),
                 tilt_angles: ArrayLike = np.array([0., 0., 0.]),
                 scattering_dispersion: float = 0.0,
                 properties: Optional[SurfaceProperties] = None,
                 material_in: Material = Materials.AIR,
                 material_out: Material = Materials.AIR,
                 name: Optional[str] = None):
        super().__init__(
            surface_type=surface_type,
            position=position,
            tilt_angles=tilt_angles,
            scattering_dispersion=scattering_dispersion,
            properties=properties,
            material_in=material_in,
            material_out=material_out,
            name=name
        )
        self.half_aperture = half_aperture
        self.curvature = curvature
        self.conic_constant = conic_constant
        self.aspheric_coefficients = np.asarray(aspheric_coefficients)
        self.central_hole_half_aperture = central_hole_half_aperture
        self.central_hole_shape = central_hole_shape
        self.aperture_shape = aperture_shape
        self.is_fresnel = is_fresnel
        self.offset = np.array([0.,0.])
        """Treat the surface as a segment centered at these aperture coordinates."""
        self._shape = SurfaceShape.ASPHERICAL

    def _validate(self):
        """Validates the internal consistency of the surface parameters."""
        name_str = "-".join([self.name,""]) if self.name is not None else ""
        if not self.half_aperture > 0:
            raise ValueError("".join([name_str, "Half-aperture must be positive."]))
        if not self.central_hole_half_aperture >= 0:
            raise ValueError("".join([name_str, "Half-aperture of the central hole cannot be negative."]))
        if not self.central_hole_half_aperture < self.half_aperture:
            raise ValueError("".join([name_str, "Half-aperture of the central must be less than surface half-aperture."]))
        if not self.aspheric_coefficients.ndim == 1:
            raise ValueError("".join([name_str, "Aspheric coefficients must be a 1D array."]))

    @property
    def aspheric_coefficients(self) -> np.ndarray:
        return self._aspheric_coefficients

    @aspheric_coefficients.setter
    def aspheric_coefficients(self, value: np.ndarray):
        if not isinstance(value, np.ndarray):
            value = np.asarray(value, dtype=np.float64)
        if len(value) > 10:
            raise(ValueError("Aspheric surfaces can be defined by a maximum of 10 aspheric coefficients."))
        self._aspheric_coefficients = value

    def sagitta(self, r: np.ndarray) -> np.ndarray:
        """Calculate the sagitta of the aspheric surface at a given radial distance in the surface refrence frame.

        Parameters
        ----------
        r : np.ndarray, list or scalar
            Radial distance(s) from the optical axis.

        Returns
        -------
        np.ndarray
            Sagitta value(s) at the given radial distance(s).
        """

        return sag(r, self.curvature, self.conic_constant, self.aspheric_coefficients, 0., self.is_fresnel)

    def get_surface_normal_vector(self, x: float, y: float) -> np.ndarray:
        """Calculate the upward-going normal unit vector at a given surface point (x,y) in the surface reference frame.

        Parameters
        ----------
        x : float
            x coordinate of the point.
        y : float
            y coordinate of the point.

        Returns
        -------
        np.ndarray
            Normal unit vector.
        """
        n = np.array([0,0,1.])

        # Flat
        if (np.abs(self.curvature) < 1e-10): 
            return n
        
        r2 = x*x+y*y

        # Aspheric component
        tot_aspher = 0.
        for i in range(len(self.aspheric_coefficients)):
            tot_aspher += 2 * (i+1) * self.aspheric_coefficients[i] * np.power(r2, i)

        arg_sqrt = 1-(1.0 + self.conic_constant) * self.curvature*self.curvature * r2
        if arg_sqrt < 0:
            raise(ValueError('Surface not well defined at the given point.'))
        
        fact = self.curvature / np.sqrt(arg_sqrt) + tot_aspher

        n[0] = - fact * x
        n[1] = - fact * y
        norm = np.sqrt(n[0]*n[0]+n[1]*n[1]+1.)

        return n/norm
    
    def _table_data(self) -> list[list[Any]]:
        """Helper function to generate data for table representation."""
        table = []
        if self.name is not None:
            table.append(["Name", self.name])
        table += [
            ["Half-Aperture", f'{self.half_aperture:.2f}'],
            ["Curvature", f'{self.curvature:.9f}'],
            ["Conic Constant", f'{self.conic_constant:.2f}'],
            ["# Aspheric Coefficients", len(self.aspheric_coefficients)],
            ["Central Hole Half-Aperture", f'{self.central_hole_half_aperture:.2f}'],
            ["Aperture Shape", self.aperture_shape],
            ["Central Hole Shape", self.central_hole_shape],
            ["Is Fresnel", self.is_fresnel],
            ["Surface Type", self.type],
            ["Position", ', '.join([f'{x:.2f}' for x in self.position])],
            ["Tilt Angles", ', '.join([f'{x:.2f}' for x in self.tilt_angles])],
            ["Scattering Sigma", f'{self.scattering_dispersion:.3f}'],
            ["Properties wavelegth", "set" if self.properties.wavelength is not None else "not set"],
            ["Properties incidence angle", "set" if self.properties.incidence_angle is not None else "not set"],
            ["Transmittance value", "set" if self.properties.transmittance is not None else "not set"],
            ["Reflectance value", "set" if self.properties.reflectance is not None else "not set"],
            ["Absorption value", "set" if self.properties.absorption is not None else "not set"],
        ]
        return table

class FlatSurface(AsphericalSurface):
    """Represents a flat optical surface.

        Parameters
        ----------
        half_aperture : float
            Half-aperture of the surface.
        central_hole_half_aperture : float, optional
            Half-aperture of the central hole (0.0 if no hole), by default 0.0
        aperture_shape : ApertureShape, optional
            Shape of the aperture (circular, hexagonal or rectangular), by default ApertureShape.CIRCULAR
        central_hole_shape : ApertureShape, optional
            Shape of the central hole (circular, hexagonal or rectangular), by default ApertureShape.CIRCULAR
        surface_type : SurfaceType, optional
            Type of surface (reflective, refractive, opaque, dummy), by default SurfaceType.REFLECTIVE
        position : array-like, optional
            Position of the surface center [x, y, z], by default None
        tilt_angles : array-like, optional
            Tilt angles (clockwise x-tilt, y-tilt and z_tild) in degrees, by default no tilt.
        scattering_dispersion : float, optional
            Standard deviation of Gaussian scattering after reflection/transmission, by default 0.0
        properties : SurfaceProperties, optional
            Reflectance/transmittance properties of the surface, by default None
        material_in : material, optional
            Material of the space before the surface front side. I.e. the material in which
            photons arriving with a negative z-component of their direction vector are travelling, 
            by default standard air.
        material_out : material, optional
            Material of the space after the surface back side. I.e. the material in which
            photons arriving with a positive z-component of their direction vector are travelling, 
            by default standard air.
        name : str, optional
            Surface name
    """

    def __init__(self,
                 half_aperture: float,
                 central_hole_half_aperture: float = 0.0,
                 aperture_shape: ApertureShape = ApertureShape.CIRCULAR,
                 central_hole_shape: ApertureShape = ApertureShape.CIRCULAR,
                 surface_type: SurfaceType = SurfaceType.REFLECTIVE,
                 position: ArrayLike = np.array([0., 0., 0.]),
                 tilt_angles: ArrayLike = np.array([0., 0., 0.]),
                 scattering_dispersion: float = 0.0,
                 properties: Optional[SurfaceProperties] = None,
                 material_in: Material = Materials.AIR,
                 material_out: Material = Materials.AIR,
                 name: Optional[str] = None):
        super().__init__(
            half_aperture=half_aperture,
            curvature=0,
            conic_constant=0,
            aspheric_coefficients=[0.],
            central_hole_half_aperture=central_hole_half_aperture,
            aperture_shape=aperture_shape,
            central_hole_shape=central_hole_shape,
            is_fresnel=False,
            surface_type=surface_type,
            position=position,
            tilt_angles=tilt_angles,
            scattering_dispersion=scattering_dispersion,
            properties=properties,
            material_in=material_in,
            material_out=material_out,
            name=name
        )
        self._shape = SurfaceShape.FLAT

class SphericalSurface(AsphericalSurface):
    """Represents a spherical optical surface.

        Parameters
        ----------
        half_aperture : float
            Half-aperture of the surface.
        curvature : float
            Surface curvature (1/radius of curvature)
        central_hole_half_aperture : float, optional
            Half-aperture of the central hole (0.0 if no hole), by default 0.0
        aperture_shape : ApertureShape, optional
            Shape of the aperture (circular, hexagonal or rectangular), by default ApertureShape.CIRCULAR
        central_hole_shape : ApertureShape, optional
            Shape of the central hole (circular, hexagonal or rectangular), by default ApertureShape.CIRCULAR
        is_fresnel : bool, optional
            Whether the surface is an ideal Fresnel lens/mirror, by default False
        surface_type : SurfaceType, optional
            Type of surface (reflective, refractive, opaque, dummy), by default SurfaceType.REFLECTIVE
        position : array-like, optional
            Position of the surface center [x, y, z], by default None
        tilt_angles : array-like, optional
            Tilt angles (clockwise x-tilt, y-tilt and z_tild) in degrees, by default no tilt.
        scattering_dispersion : float, optional
            Standard deviation of Gaussian scattering after reflection/transmission, by default 0.
        properties : SurfaceProperties, optional
            Reflectance/transmittance properties of the surface, by default None
        material_in : material, optional
            Material of the space before the surface front side. I.e. the material in which
            photons arriving with a negative z-component of their direction vector are travelling, 
            by default standard air.
        material_out : material, optional
            Material of the space after the surface back side. I.e. the material in which
            photons arriving with a positive z-component of their direction vector are travelling, 
            by default standard air.
        name : str, optional
            Surface name
    """

    def __init__(self,
                 half_aperture: float,
                 curvature: float,
                 central_hole_half_aperture: float = 0.,
                 aperture_shape: ApertureShape = ApertureShape.CIRCULAR,
                 central_hole_shape: ApertureShape = ApertureShape.CIRCULAR,
                 is_fresnel: bool = False,
                 surface_type: SurfaceType = SurfaceType.REFLECTIVE,
                 position: ArrayLike = np.array([0., 0., 0.]),
                 tilt_angles: ArrayLike = np.array([0., 0., 0.]),
                 scattering_dispersion: float = 0.,
                 properties: Optional[SurfaceProperties] = None,
                 material_in: Material = Materials.AIR,
                 material_out: Material = Materials.AIR,
                 name: Optional[str] = None):
        super().__init__(
            half_aperture=half_aperture,
            curvature=curvature,
            conic_constant=0,
            aspheric_coefficients=[0.],
            central_hole_half_aperture=central_hole_half_aperture,
            aperture_shape=aperture_shape,
            central_hole_shape=central_hole_shape,
            is_fresnel=is_fresnel,
            surface_type=surface_type,
            position=position,
            tilt_angles=tilt_angles,
            scattering_dispersion=scattering_dispersion,
            properties=properties,
            material_in=material_in,
            material_out=material_out,
            name=name
        )
        self._shape = SurfaceShape.SPHERICAL


class CylindricalSurface(Surface):
    """Represents a cylindrical optical surface.

        Parameters
        ----------
        radius : float
            Radius of the cylinder.
        height : float
            Height of the cylinder.
        has_top : bool, optional
            Whether the cylinder has a top cap, by default True.
        has_bottom : bool, optional
            Whether the cylinder has a bottom cap, by default True.
        position : array-like, optional
            Position of cylinder center [x, y, z] (half-height on the cylinder axis), by default None.
        tilt_angles : array-like, optional
            Tilt angles (clockwise x-tilt, y-tilt and z_tild) in degrees, by default no tilt.
        surface_type : SurfaceType, optional
            Type of surface (opaque or sensitive), by default SurfaceType.OPAQUE.
        name : str, optional
            Surface name
    """

    def __init__(self,
                 radius: float,
                 height: float,
                 has_top: bool = True,
                 has_bottom: bool = True,
                 position: ArrayLike = np.array([0., 0., 0.]),
                 tilt_angles: ArrayLike = np.array([0., 0., 0.]),
                 surface_type: SurfaceType = SurfaceType.OPAQUE,
                 name: Optional[str] = None):

        scattering_dispersion = 0.
        properties = None
        material_in = Materials.AIR
        material_out = Materials.AIR
        super().__init__(
            surface_type=surface_type,
            position=position,
            tilt_angles=tilt_angles,
            scattering_dispersion=scattering_dispersion,
            properties=properties,
            material_in=material_in,
            material_out=material_out,
            name=name
        )
        self.radius = radius
        self.height = height
        self.top = has_top
        self.bottom = has_bottom
        self._shape = SurfaceShape.CYLINDRICAL
    
    def _validate(self):
        """Validates the internal consistency of the surface parameters."""
        name_str = "-".join([self.name,""]) if self.name is not None else ""
        if not self.radius > 0:
            raise ValueError("".join([name_str, "Cylinder radius must be positive."]))
        if not self.height > 0:
            raise ValueError("".join([name_str, "Cylinder height must be positive."]))
    
    def _table_data(self) -> list[list[Any]]:
        """Helper function to generate data for table representation."""
        table = []
        if self.name is not None:
            table.append(["Name", self.name])
        table += [
            ["Radius", f'{self.radius:.2f}'],
            ["Height", f'{self.height:.2f}'],
            ["Has top", self.top],
            ["Has bottom", self.bottom],
            ["Surface Type", self.type],
            ["Position", ', '.join([f'{x:.2f}' for x in self.position])],
            ["Tilt Angles", ', '.join([f'{x:.2f}' for x in self.tilt_angles])],
        ]
        return table


class SipmTileSurface(FlatSurface):
    """Represents a SiPM tile of NxN pixels.

        Parameters
        ----------
        pixels_per_side : int
            Number of pixels per side.
        pixel_active_side : float
            Side length of the active area of a single pixel.
        pixels_separation : float
            Separation between pixels.
        border_to_active_area : float
            Width of the border from the edge of the tile to the active area.
        position : array-like, optional
            Position of the surface center [x, y, z], by default None
        tilt_angles : array-like, optional
            Tilt angles (clockwise x-tilt, y-tilt and z_tild) in degrees, by default no tilt.
        scattering_dispersion : float, optional
            Standard deviation of Gaussian scattering after reflection/transmission, by default 0.0
        properties : SurfaceProperties, optional
            Reflectance/transmittance properties of the surface, by default None
        material_in : material, optional
            Material of the space before the surface front side. I.e. the material in which
            photons arriving with a negative z-component of their direction vector are travelling, 
            by default standard air.
        material_out : material, optional
            Material of the space after the surface back side. I.e. the material in which
            photons arriving with a positive z-component of their direction vector are travelling, 
            by default standard air.
        name : str, optional
            Surface name
        
        Notes
        -----
        Tile layout::

              y       pixels_separation
              ^    ________ <--> ________
              |   |        |    |        |
              |   |        |    |        |
              |   |        |    |        |
              |   |________|    |________|
              |<->               <------>
              | |            pixel_active_side
              | border_to_active_area
            (0,0)---------> x

    """

    def __init__(self,
                 pixels_per_side: int,
                 pixel_active_side: float,
                 pixels_separation: float,
                 border_to_active_area: float,
                 position: ArrayLike = np.array([0., 0., 0.]),
                 tilt_angles: ArrayLike = np.array([0., 0., 0.]),
                 scattering_dispersion: float = 0.0,
                 surface_type: SurfaceType = SurfaceType.SENSITIVE,
                 properties: Optional[SurfaceProperties] = None,
                 material_in: Material = Materials.AIR,
                 material_out: Material = Materials.AIR,
                 name: Optional[str] = None
                ):
        
        # Compute tile size
        tile_side = pixels_per_side * pixel_active_side + (pixels_per_side - 1) * pixels_separation
        half_aperture = 0.5 * tile_side + border_to_active_area

        super().__init__(
            half_aperture = half_aperture,
            central_hole_half_aperture = 0.0,
            aperture_shape = ApertureShape.SQUARE,
            surface_type = SurfaceType.SENSITIVE_BACK,
            position = position,
            tilt_angles = tilt_angles,
            scattering_dispersion = scattering_dispersion,
            properties = properties,
            material_in = material_in,
            material_out = material_out,
            name = name
        )
        self.type = surface_type
        self._pixel_active_side = pixel_active_side
        self._pixels_separation = pixels_separation
        self._border_to_active_area = border_to_active_area
        self._pixels_per_side = pixels_per_side
        self._n_pixels = pixels_per_side * pixels_per_side

        self.sensor_type = SensorSurfaceType.SIPM_TILE
        self._shape = SurfaceShape.FLAT
        self.sensor_id = 0
    
    @property
    def type(self) -> SurfaceType:
        return self._type

    @type.setter
    def type(self, value: SurfaceType):
        sensitive_list = [
            SurfaceType.SENSITIVE_BACK,
            SurfaceType.SENSITIVE_FRONT,
            SurfaceType.SENSITIVE,
            SurfaceType.REFLECTIVE_SENSITIVE,
        ]
        if value not in sensitive_list:
            raise(ValueError("SipmTileSurface can be only a sensitive surface."))
        self._type = value
    
    @property
    def pixels_per_side(self) -> int:
        return self._pixels_per_side

    @pixels_per_side.setter
    def pixels_per_side(self, value: int):
        self._pixels_per_side = value
        self._n_pixels = value * value
        self._update_size()
    
    @property
    def pixel_active_side(self) -> float:
        return self._pixel_active_side
    
    @pixel_active_side.setter
    def pixel_active_side(self, value: float):
        self._pixel_active_side = value
        self._update_size()
    
    @property
    def pixels_separation(self) -> float:
        return self._pixels_separation
    
    @pixels_separation.setter
    def pixels_separation(self, value: float):
        self._pixels_separation = value
        self._update_size()
    
    @property
    def border_to_active_area(self) -> float:
        return self._border_to_active_area
    
    @border_to_active_area.setter
    def border_to_active_area(self, value: float):
        self._border_to_active_area = value
        self._update_size()
    
    def _update_size(self):
        tile_side = self.pixels_per_side * self.pixel_active_side + (self.pixels_per_side - 1) * self.pixels_separation
        self.half_aperture = 0.5 * tile_side + self.border_to_active_area


# Copyright (C) 2024- Davide Mollica <davide.mollica@inaf.it>
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of iactsim.
#
# iactsim is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# iactsim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with iactsim.  If not, see <https://www.gnu.org/licenses/>.

def is_notebook() -> bool:
    """
    Checks if the code is running in a Jupyter Notebook.

    Returns
    -------
    bool
        True if running in a Jupyter Notebook, False otherwise.
    """
    try:
        from IPython import get_ipython
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return True 
        else:
            return False
    except (NameError, ImportError):
        return False


# Copyright (C) 2024- Davide Mollica <davide.mollica@inaf.it>
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of iactsim.
#
# iactsim is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# iactsim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with iactsim.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import cupy as cp

from ..visualization._iactsim_style import iactsim_style

class Waveform:
    """
    Represents the normalized output signal of the electronics in response to a single photon 
    detection as a function of the time from its arrival time.
    Is assumed to be 0 outside the defined time window.

    Parameters
    ----------
    time : numpy.ndarray
        Array of time values corresponding where the waveform is defined, must be monotonically increasing.
    amplitude : numpy.ndarray
        Array of waveform amplitudes corresponding to the time array.
    coupling : str
        Coupling of the signal. Must be "AC" or "DC" (case insensitive).
        It is only used when calling `normalize()` method without a normalization value:

          - "AC": normalize by the waveform peak;
          - "DC": normalize by the waveform integral.
    
    Notes
    -----
        A "dummy" waveform can be created using the default `None` arguments:

        .. code-block::`pyhton`

            dummy_wave = Waveform()
        

    Raises
    ------
    TypeError
        If the input argument are not Numpy ndarray.
    ValueError
        If the input argument have more than one dimension.
    ValueError
        If the input argument do not have the same length. 
    ValueError
        If the time array is not monotonically increasing. 
    ValueError
        If the time array is not equispaced. 
    ValueError
        If the signal coupling is not valid.
    """

    def __init__(self, time: np.ndarray = None, amplitude: np.ndarray = None, coupling: str = None):
        
        if all([x is None for x in [time, amplitude, coupling]]):
            self._time = None
            self._amplitude = None
            self._coupling = None
            return

        if not isinstance(time, np.ndarray) or not isinstance(amplitude, np.ndarray):
            raise TypeError("time and amplitude must be NumPy arrays")
        
        if time.ndim != 1 or amplitude.ndim != 1:
            raise ValueError("time and amplitude must be 1D arrays")
        
        if len(time) != len(amplitude):
            raise ValueError("time and amplitude arrays must have the same length")

        if not np.all(np.diff(time) > 0):
            raise ValueError("time array must be monotonically increasing")

        if not isinstance(coupling, str) or coupling.lower() not in ['ac', 'dc']:
            raise(ValueError(f"Invalid signal coupling '{coupling}'. Accepted amplitudes are 'AC' and 'DC' (case insensitive)."))

        time_diffs = np.diff(time)
        tolerance = 1e-3 
        if not np.allclose(time_diffs, time_diffs[0], atol=tolerance):
            raise ValueError("time array must be equispaced")

        self._time = time
        self._amplitude = amplitude
        self._coupling = coupling.lower()
        self._extent = self._time[-1] - self.time[0]
        self.normalize()

    @property
    def time(self):
        """
        Returns the time array.
        """
        return self._time
    
    @property
    def amplitude(self):
        """
        Returns the amplitude array.
        """
        return self._amplitude
    
    @property
    def coupling(self):
        """
        Returns the signal coupling.
        """
        return self._coupling
    
    def get_peak_time(self) -> float:
        """
        Returns the time corresponding to the peak amplitude of the waveform.
        """
        peak_index = np.argmax(self._amplitude)
        return self._time[peak_index]

    def get_peak_amplitude(self) -> float:
        """
        Returns the peak amplitude of the waveform.
        """
        return np.max(self._amplitude)

    def get_integral(self) -> float:
        """
        Returns the approximate integral of the waveform using the trapezoidal rule.
        """
        return np.trapezoid(self._amplitude, x=self._time)

    def get_squared_integral(self) -> float:
        """
        Returns the approximate integral of the squared waveform using the trapezoidal rule.
        """
        return np.trapezoid(self._amplitude**2, x=self._time)
    
    def get_size(self) -> int:
        """
        Returns the waveform number of time bins.
        """
        if len(self._time) != len(self._amplitude):
            raise(RuntimeError("The number of amplitude amplitudes must be equal to the number of time-bins"))
        return len(self._time)
    
    def get_extent(self) -> float:
        """
        Returns the waveform time-extent.
        """
        return self._extent
    
    def normalize(self, norm=None):
        """Normalize the waveform amplitude.

        Parameters
        ----------
        norm : float, optional
            Normalization factor. If None, it is determined based on the
            coupling type. For 'AC' coupling, the waveform is normalized to
            the peak amplitude. For 'DC' coupling it's normalized
            to the integral. By default, None.
        """
        if norm is None:
            if self.coupling == 'ac':
                norm = 1./self.get_peak_amplitude()
            else:
                norm = 1./self.get_integral()
        self._amplitude *= norm

    @iactsim_style
    def plot(self):
        if self.time is None or self.amplitude is None:
            raise(RuntimeError("Waveform has not been initialized."))
        
        import matplotlib.pyplot as plt
        fig, ax = plt.subplots()
        plt.plot(self.time, self.amplitude, color='black')
        plt.xlabel('Time (ns)')
        plt.ylabel('Signal amplitude')
        ax.grid(which='both')
        return fig, ax

    def get_waveform_texture(self):
        # Texture Setup
        width = len(self.time)
        texture_data = self.amplitude.astype(np.float32)
        
        # Create a CUDAarray for the 1D texture
        desc = cp.cuda.texture.ChannelFormatDescriptor(32, 0, 0, 0, cp.cuda.runtime.cudaChannelFormatKindFloat)
        cu_array = cp.cuda.texture.CUDAarray(desc, width)
        cu_array.copy_from(texture_data)
        
        # Create another texture object (Linear Interpolation)
        res_desc_linear = cp.cuda.texture.ResourceDescriptor(
        cp.cuda.runtime.cudaResourceTypeArray,
        cuArr=cu_array
        )
        tex_desc_linear = cp.cuda.texture.TextureDescriptor(
            addressModes = (cp.cuda.runtime.cudaAddressModeClamp,),
            filterMode = cp.cuda.runtime.cudaFilterModeLinear,
            readMode = cp.cuda.runtime.cudaReadModeElementType,
            normalizedCoords = False
        )
        return cp.cuda.texture.TextureObject(res_desc_linear, tex_desc_linear), 1./(self.time[1]-self.time[0]), self.time[0] 


import pytest
import cupy as cp
import iactsim
import numpy as np

@pytest.fixture(scope="module")
def telescope_setup():
    """Initializes the telescope and GPU context."""
    astri = iactsim.models.astri.AstriTelescope()

    m1 = astri.optical_system['M1']

    offsets = np.flip(np.asarray([
        [856.5, 0.0],
        [428.2, 741.4],
        [-428.2, 741.4],
        [-856.5, 0.0],
        [-428.2, -741.4],
        [428.2, -741.4],
        [1280.5, 738.8],
        [0.0, 1478.6],
        [-1280.5, 739.3],
        [-1280.5, -739.3],
        [0.0, -1478.6],
        [1280.5, -739.3],
        [1704.8, 0.0],
        [852.4, 1476.4],
        [-852.4, 1476.4],
        [-1704.8, 0.0],
        [-852.4, -1476.4],
        [852.4, -1476.4],
    ]), axis=1)

    offsets[:,0]*=-1

    for i,offset in enumerate(offsets):
        segment = iactsim.optics.AsphericalSurface(
            conic_constant=m1.conic_constant,
            aspheric_coefficients=m1.aspheric_coefficients,
            half_aperture=846./2., 
            curvature=m1.curvature,
            position=[offset[0], offset[1], m1.sagitta(np.sqrt(offset[0]**2+offset[1]**2))],
            surface_type=iactsim.optics.SurfaceType.REFLECTIVE_FRONT,
            name = f'Mirror{i}',
            aperture_shape=iactsim.optics.ApertureShape.HEXAGONAL
        )
        segment.offset = offset
        astri.optical_system.add_surface(segment, replace=True)
                    
    astri.optical_system.remove_surface('M1')

    astri.optical_system.camera_window_parameters = (240, 8.61, 1.5, 1, 3) 

    astri.optical_system.build_masts()

    astri.cuda_init()

    return astri

def test_ray_tracing_performance_astri(benchmark, telescope_setup):
    telescope = telescope_setup
    source = iactsim.optics.sources.Source(telescope)
    source.positions.radial_uniformity = False
    source.positions.random = True
    source.positions.r_max = 2500
    source.set_target(distance=11000)

    def setup_round_on_axis():
        ps, vs, wls, ts = source.generate(1_000_000)
        cp.cuda.Device().synchronize()
        return ((telescope, ps, vs, wls, ts), {})
    
    def run_trace(telescope, ps, vs, wls, ts):
        telescope.trace_photons(ps, vs, wls, ts)
        cp.cuda.Device().synchronize()
    
    benchmark.pedantic(run_trace, setup=setup_round_on_axis, rounds=100, iterations=1)

def test_ray_tracing_performance_astri_off_axis(benchmark, telescope_setup):
    telescope = telescope_setup
    source = iactsim.optics.sources.Source(telescope)
    source.positions.radial_uniformity = False
    source.positions.random = True
    source.positions.r_max = 2500
    source.directions.altitude += 1
    source.set_target(distance=11000)
    def setup_round_off_axis():
        ps, vs, wls, ts = source.generate(1_000_000)
        cp.cuda.Device().synchronize()
        return ((telescope, ps, vs, wls, ts), {})
    
    def run_trace(telescope, ps, vs, wls, ts):
        telescope.trace_photons(ps, vs, wls, ts)
        cp.cuda.Device().synchronize()
    
    benchmark.pedantic(run_trace, setup=setup_round_off_axis, rounds=100, iterations=1)

def test_ray_tracing_performance_astri_pdms(benchmark, telescope_setup):
    telescope = telescope_setup
    telescope.optical_system.build_camera_modules()
    telescope.cuda_init()
    source = iactsim.optics.sources.Source(telescope)
    source.positions.radial_uniformity = False
    source.positions.random = True
    source.positions.r_max = 2500
    source.directions.altitude += 1
    source.set_target(distance=11000)

    def setup_round_pdms():
        ps, vs, wls, ts = source.generate(1_000_000)
        cp.cuda.Device().synchronize()
        return ((telescope, ps, vs, wls, ts), {})
    
    def run_trace(telescope, ps, vs, wls, ts):
        telescope.trace_photons(ps, vs, wls, ts)
        cp.cuda.Device().synchronize()
    
    benchmark.pedantic(run_trace, setup=setup_round_pdms, rounds=100, iterations=1)

def test_ray_tracing_performance_astri_pdms_window(benchmark, telescope_setup):
    telescope = telescope_setup
    telescope.optical_system.build_camera_modules()
    telescope.optical_system.camera_window_parameters = (240, 8.61, 1.5, 1, 3) 
    telescope.cuda_init()
    source = iactsim.optics.sources.Source(telescope)
    source.positions.radial_uniformity = False
    source.positions.random = True
    source.positions.r_max = 2500
    source.directions.altitude += 1
    source.set_target(distance=11000)

    def setup_round_pdms_window():
        ps, vs, wls, ts = source.generate(1_000_000)
        cp.cuda.Device().synchronize()
        return ((telescope, ps, vs, wls, ts), {})
    
    def run_trace(telescope, ps, vs, wls, ts):
        telescope.trace_photons(ps, vs, wls, ts)
        cp.cuda.Device().synchronize()
    
    benchmark.pedantic(run_trace, setup=setup_round_pdms_window, rounds=100, iterations=1)

def test_ray_tracing_performance_astri_pdms_window_props(benchmark, telescope_setup):
    telescope = telescope_setup

    telescope.optical_system.build_camera_modules()

    telescope.optical_system.camera_window_parameters = (240, 8.61, 1.5, 1, 3) 

    gen = iactsim.optics.FresnelSurfacePropertiesGenerator()
    wls = np.arange(200,1001,1.)
    angls = np.arange(0,91,1.)
    gen.generate(telescope.optical_system['WindowLayer1'], wls, angls,inplace=True)
    gen.generate(telescope.optical_system['WindowLayer2'], wls, angls,inplace=True)
    gen.generate(telescope.optical_system['WindowLayer3'], wls, angls,inplace=True)
    telescope.cuda_init()

    source = iactsim.optics.sources.Source(telescope)
    source.positions.radial_uniformity = False
    source.positions.random = True
    source.positions.r_max = 2500
    source.directions.altitude += 1
    source.set_target(distance=11000)

    def setup_round_pdms_window():
        ps, vs, wls, ts = source.generate(1_000_000)
        cp.cuda.Device().synchronize()
        return ((telescope, ps, vs, wls, ts), {})
    
    def run_trace(telescope, ps, vs, wls, ts):
        telescope.trace_photons(ps, vs, wls, ts)
        cp.cuda.Device().synchronize()
    
    benchmark.pedantic(run_trace, setup=setup_round_pdms_window, rounds=100, iterations=1)

def test_ray_tracing_performance_astri_pdms_props_window_props(benchmark, telescope_setup):
    # SiPM optical properties
    sipm = iactsim.optics.SipmStackSimulator(
        layers=["SiO2", "ZrO2"],
        thicknesses=[3250e-9, 35e-9],
        medium="Air",
        type_layers=['c', 'c'],
        depletion_layer_depth=3.,
        depletion_layer_width=2600,
        fill_factor=0.75
    )
    sipm.set_simulation_params(
        wl_range=np.arange(200,1000,1),
        angle_range=np.arange(0,91,1)
    )
    sipm.run('u')
    sipm.constrain_pde(462, 0.582517*0.97)
    sipm_surface_prop = sipm.get_surfcace_properties_obj()

    telescope = telescope_setup

    telescope.optical_system.build_camera_modules()

    telescope.optical_system.camera_window_parameters = (240, 8.61, 1.5, 1, 3) 

    gen = iactsim.optics.FresnelSurfacePropertiesGenerator()
    wls = np.arange(200,1001,1.)
    angls = np.arange(0,91,1.)
    gen.generate(telescope.optical_system['WindowLayer1'], wls, angls,inplace=True)
    gen.generate(telescope.optical_system['WindowLayer2'], wls, angls,inplace=True)
    gen.generate(telescope.optical_system['WindowLayer3'], wls, angls,inplace=True)

    for s in telescope.optical_system:
        if s.name.startswith('PDM'):
            s.properties = sipm_surface_prop

    telescope.cuda_init()

    source = iactsim.optics.sources.Source(telescope)
    source.positions.radial_uniformity = False
    source.positions.random = True
    source.positions.r_max = 2500
    source.directions.altitude += 1
    source.set_target(distance=11000)
    
    def setup_round_pdms_window():
        ps, vs, wls, ts = source.generate(1_000_000)
        cp.cuda.Device().synchronize()
        return ((telescope, ps, vs, wls, ts), {})
    
    def run_trace(telescope, ps, vs, wls, ts):
        telescope.trace_photons(ps, vs, wls, ts)
        cp.cuda.Device().synchronize()
    
    benchmark.pedantic(run_trace, setup=setup_round_pdms_window, rounds=100, iterations=1)


# Copyright (C) 2024- Davide Mollica <davide.mollica@inaf.it>
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of iactsim.
#
# iactsim is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# iactsim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with iactsim.  If not, see <https://www.gnu.org/licenses/>.

import cupy as cp
import numpy as np

from abc import ABC, abstractmethod

from .._utils import _disk_picking

def check_position(position, dtype=cp.float32):
    # Convert position to a CuPy array if it's not already
    
    if isinstance(position, (list, tuple)):
        position = np.array(position)
    
    if isinstance(position, np.ndarray):
        position = cp.asarray(position, dtype=dtype)
    
    if not isinstance(position, cp.ndarray):
        raise TypeError("position must be a list, tuple, Numpy ndarray or CuPy ndarray.")

    if position.shape != (3,):
        raise ValueError("position must be a 3-element array_like object.")
    
    return position

class AbstractPhotonPositions(ABC):
    """
    Abstract base class for a photon position generator.

    This class defines the interface for generating photon positions in 3D space.

    Attributes
    ----------
    position : cp.ndarray
        The (x, y, z) coordinates of the point or center of the photon source.
    rotation : cp.ndarray
        Rotation to be applied to the generated points.
    dtype : type
        The desired data type of the returned array.
    seed : int | None
        Seed for the random number generator.
    """

    def __init__(self, position: tuple | list | np.ndarray | cp.ndarray, rotation: np.ndarray | cp.ndarray = None, dtype: type = cp.float32, seed: int | None = None):
        self.position = position
        self.dtype = dtype
        self.seed = seed
        self.rotation = None

    @property
    def position(self):
        return self._position

    @position.setter
    def position(self, a_position):
        self._position = check_position(a_position)

    @property
    def rotation(self):
        return self._rotation

    @rotation.setter
    def rotation(self, a_rotation):
        if a_rotation is not None:
            a_rotation = cp.asarray(a_rotation, self.dtype)
        self._rotation = a_rotation
    
    @property
    def seed(self):
        return self._seed
    
    @seed.setter
    def seed(self, a_seed):
        self.rng = cp.random.default_rng(cp.random.Philox4x3210(a_seed))
        self._seed = a_seed
    
    @abstractmethod
    def _generate(self, n: int) -> cp.ndarray:
        if not isinstance(n, int) or n <= 0:
            raise ValueError("`n` must be a positive integer.")
        pass

    def generate(self, n: int) -> cp.ndarray:
        """
        Generates photon positions.

        Parameters
        ----------
        n : int
            Number of photon positions to generate.

        Returns
        -------
        positions : cp.ndarray
            A CuPy array of shape (n, 3) where each row represents a
            photon position vector (x, y, z).

        Raises
        ------
        ValueError
            If n is not a positive integer.
        """
        ps = self._generate(n)
        if self.rotation is not None:
            ps = ps @ self.rotation.T
        ps += self.position
       
        return ps 

class PointLike(AbstractPhotonPositions):
    """
    Represents a point source by generating a set of identical 3D points.

    This class creates an array of `n` identical 3D points, all located at the
    specified `position`.

    Parameters
    ----------
    position : tuple, list, or cp.ndarray, shape (3,)
        The (x, y, z) coordinates of the point source. Can be a tuple, list, or CuPy array.
    dtype : type, optional
        The desired data type of the returned array. Defaults to `cp.float32`.
    seed : int | None, optional
        Seed for the random number generator.

    Raises
    ------
    ValueError
        If `position` is not a 3-element tuple, list, or CuPy array.
    """

    def __init__(self, position: tuple | list | cp.ndarray, rotation: cp.ndarray = None, dtype: type = cp.float32, seed: int | None = None):
        super().__init__(position, rotation=rotation, dtype=dtype, seed=seed)

    def _generate(self, n: int) -> cp.ndarray:
        """
        Generates the set of identical 3D points representing the point source.

        Parameters
        ----------
        n : int
            The number of identical points to generate.

        Returns
        -------
        ps : cp.ndarray, shape (n, 3)
            A CuPy array where each row represents a point at the given `position`.

        Raises
        ------
        ValueError
            If `n` is not a positive integer.
        """
        if not isinstance(n, int) or n <= 0:
            raise ValueError("`n` must be a positive integer.")
        
        ps = cp.zeros((n, 3), dtype=self.dtype)

        return ps

class UniformDisk(AbstractPhotonPositions):
    """
    Generates uniformly distributed points on a disk (or annulus) parallel to the x-y plane.

    Parameters
    ----------
    r_min : float
        Minimum radius of the disk (or annulus). If `r_min` > 0, it defines the inner radius of an annulus.
    r_max : float
        Maximum radius of the disk (or annulus). Defines the outer radius.
    position : array_like, shape (3,)
        Center of the disk (or annulus) in Cartesian coordinates. Can be a list, tuple, NumPy array or CuPy array.
    radial_uniformity : bool, optional
        If True (default), uses a radial uniform distribution. If False, uses a uniform distribution in area.
    random : bool, optional
        If True (default), generates random points. If False, generates a grid of points.
    seed : int | None, optional
        A seed to initialize the random number generator.
    dtype : type, optional
        Desired data type of the output array, by default cp.float32.

    Raises
    ------
    ValueError
        If `position` does not have a length of 3.
        If `r_min` is not >= 0 and less than `r_max`.
    TypeError
        If `position` is not a list, tuple, NumPy array or CuPy array.
    """

    def __init__(self,
                 r_min: float,
                 r_max: float,
                 position: list | tuple | np.ndarray | cp.ndarray = cp.asarray([0., 0., 0.]),
                 rotation: np.ndarray | cp.ndarray = None,
                 radial_uniformity: bool = False,
                 random: bool = True,
                 seed: int | None = None,
                 dtype: type = cp.float32):
        super().__init__(position, rotation=rotation, dtype=dtype, seed=seed)
        self.r_min = r_min 
        self.r_max = r_max
        self.radial_uniformity = radial_uniformity
        self.random = random

    @property
    def r_min(self):
        return self._r_min

    @r_min.setter
    def r_min(self, value):
        if not (isinstance(value, (int, float)) and value >= 0):
            raise ValueError("r_min must be a non-negative number")
        self._r_min = float(value)

    @property
    def r_max(self):
        return self._r_max

    @r_max.setter
    def r_max(self, value):
        if value < 0:
            raise ValueError("r_max must be a positive number")
        self._r_max = float(value)

    def _check_radii(self):
        """Checks that r_min and r_max are valid."""
        if self.r_min >= self.r_max:
            raise ValueError("r_min must be strictly less than r_max")

    def _generate(self, n: int) -> cp.ndarray:
        """
        Generate points on the disk (or annulus).

        Parameters
        ----------
        n : int
            Number of points to generate.

        Returns
        -------
        ps : ndarray, shape (n, 3)
            The generated points on the disk (or annulus). Each row represents a point (x, y, z).
        """
        self._check_radii()

        if self.random:
            ps = _disk_picking(self.r_min, self.r_max, n, self.seed, radial_uniformity=self.radial_uniformity, dtype=self.dtype)
        else:
            # The number of points could be changed
            if self.radial_uniformity:
                n_theta = int(cp.sqrt(n))
                n_r = n_theta
                n = n_r*n_theta

                theta = cp.linspace(0, 2*cp.pi, n_theta, endpoint=False)
                r = cp.linspace(self.r_min, self.r_max, n_r)
                
                p1, p2 = cp.meshgrid(theta, r, sparse=False)
                p1 = p1.reshape((n,))
                p2 = p2.reshape((n,))

                ps = cp.empty((n, 3), dtype=self.dtype)
                ps[:, 0] = p2 * cp.cos(p1)
                ps[:, 1] = p2 * cp.sin(p1)
                ps[:, 2] = 0.
            
            else:
                n = n / (1. - self.r_min**2 / self.r_max**2) / cp.pi * 4
                n = int(cp.sqrt(n))

                x = cp.linspace(-self.r_max, self.r_max, n)
                p1, p2 = cp.meshgrid(x, x, sparse=False)
                p1 = p1.reshape((n * n,))
                p2 = p2.reshape((n * n,))
                r = cp.sqrt(p1**2 + p2**2)
                where = (r > self.r_min) & (r < self.r_max)
                p1 = p1[where]
                p2 = p2[where]
                p3 = cp.zeros_like(p1)

                n = p1.shape[0]
                ps = cp.empty((n, 3), dtype=self.dtype)
                ps[:, 0] = p1
                ps[:, 1] = p2
                ps[:, 2] = p3

        return ps

class UniformSphericalCap(AbstractPhotonPositions):
    """
    Generates uniformly distributed points on a spherical cap.

    The spherical cap is defined by its center (`position`), radius,
    and an angular extent (or "opening angle").
    Points are uniformly distributed over the cap's surface and are
    generated symmetrically around the z-axis, relative to the center.

    Parameters
    ----------
    position : array_like, shape (3,)
        The center of the sphere in Cartesian coordinates (x, y, z).
        Can be a list, tuple, NumPy array or CuPy array.
    radius : float
        The radius of the sphere.
    angular_extent : float
        The opening angle of the cap in degrees, measured from the positive
        z-axis. Must be between 0 and 180 degrees.
    seed : int | None, optional
        An optional seed for the random number generator. If None, the
        generator is seeded using unpredictable data from the operating system.
    dtype : type, optional
        The desired data type of the returned array (default: np.float32).

    Raises
    ------
    ValueError
        If angular_extent is not within the range [0, 180].
        If position does not have a length of 3.
    TypeError
        If position is not a list, tuple, or ndarray.
    """

    def __init__(self, 
                 radius: float,
                 angular_extent: float,
                 seed: int | None = None,
                 position: list | tuple | np.ndarray | cp.ndarray = cp.asarray([0., 0., 0.]),
                 rotation: np.ndarray | cp.ndarray = None,
                 dtype: type = cp.float32):
        super().__init__(position, rotation=rotation, dtype=dtype, seed=seed)

        if not (0 <= angular_extent <= 180):
            raise ValueError("angular_extent must be between 0 and 180 degrees")

        self.radius = radius
        self.angular_extent = angular_extent

    def _generate(self, n: int) -> cp.ndarray:
        """
        Generates `n` random points on the spherical cap.

        Parameters
        ----------
        n : int
            The number of points to generate.

        Returns
        -------
        ndarray
            A CuPy array of shape (n, 3) containing the Cartesian coordinates
            (x, y, z) of the generated points. The points are relative to the origin
            of the coordinate system, not the center of the sphere.
        """
        # Azimuth (longitude)
        theta = 2. * cp.pi * self.rng.random(n)

        # Altitude (latitude)
        cosmin = cp.cos(cp.deg2rad(self.angular_extent))
        cos = self.rng.random(n) * (1. - cosmin) + cosmin
        phi = cp.arccos(cos)

        ps = cp.empty((n, 3), dtype=self.dtype)
        sinph = cp.sin(phi)
        ps[:, 0] = self.radius * cp.cos(theta) * sinph
        ps[:, 1] = self.radius * cp.sin(theta) * sinph
        ps[:, 2] = self.radius * cp.cos(phi)

        return ps